/* this file is published under the gnu gpl license, version 3 or any later */

#define _GNU_SOURCE

#include <dlfcn.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

static const char gtkbfc_version_string[] = "gtkbfc version 0.1";

static char *helper = "/etc/gtkbfc-helper";

static GObject *(*old_constructor) (GType type, guint n_construct_properties, GObjectConstructParam *construct_properties);

static gchar *last_path;
static gchar *last_filename;

#if 1
# define DEBUG(stmt) stmt;
#else
# define DEBUG(stmt)
#endif

/////////////////////////////////////////////////////////////////////////////

gchar *gtk_file_chooser_get_filename (GtkFileChooser *chooser)
{
  DEBUG ((printf ("gtk_file_chooser_get_filename<%s>\n", last_path)))
  return g_strdup (last_path ? last_path : "/nonex1st4nt");
}

gboolean gtk_file_chooser_select_filename (GtkFileChooser *chooser, const char *filename)
{
  return 1;
}

void gtk_file_chooser_unselect_all (GtkFileChooser *chooser)
{
}

gboolean gtk_file_chooser_set_filename (GtkFileChooser *chooser, const char *filename)
{
  DEBUG ((printf ("gtk_file_chooser_set_filename<%s>\n", filename)))
  g_free (last_filename); last_filename = g_strdup (filename);
  return 1;
}

void gtk_file_chooser_set_current_name (GtkFileChooser *chooser, const char *filename)
{
  DEBUG ((printf ("gtk_file_chooser_set_current_name<%s>\n", filename)))
}

GSList *gtk_file_chooser_get_filenames (GtkFileChooser *chooser)
{
  DEBUG ((printf ("gtk_file_chooser_get_filenames\n")))
  return g_slist_prepend (0, gtk_file_chooser_get_filename (chooser));
}

gboolean gtk_file_chooser_set_current_folder (GtkFileChooser *chooser, const gchar *folder)
{
  DEBUG ((printf ("gtk_file_chooser_set_current_folder<%s>\n", folder)))
  g_free (last_filename); last_filename = 0;
  return 1;
}

gchar * gtk_file_chooser_get_current_folder (GtkFileChooser *chooser)
{
  DEBUG ((printf ("gtk_file_chooser_get_current_folder\n")))
  g_free (last_filename); last_filename = 0;
  return g_strdup ("/tmp/gtkbfc");
}

/////////////////////////////////////////////////////////////////////////////

static const gchar *
from_uri (const gchar *uri)
{
  const gchar *rv = uri;

  if (!strncmp (uri, "file:///", sizeof ("file:///") - 1))
    rv += sizeof ("file:///") - 1;

  return rv;
}

static gchar *
to_uri (gchar *file)
{
  gchar *rv = NULL;

  if (file)
    {
      GString *str = g_string_new ("file://");
      str = g_string_append (str, file);
      rv = g_strdup (str->str);
      g_string_free (str, TRUE);
      g_free (file);
    }

  return rv;
}

static GSList *
to_uris (GSList *files)
{
  GSList *item = files;

  for (; item; item = g_slist_next (item))
    {
      GString *str = g_string_new ("file://");
      gchar *cur = item->data;

      str = g_string_append (str, cur);
      item->data = g_string_free (str, FALSE);
      g_free (cur);
    }

  return files;
}

gchar *gtk_file_chooser_get_uri (GtkFileChooser *chooser)
{
  gchar *uri = to_uri(gtk_file_chooser_get_filename (chooser));
  DEBUG ((printf ("gtk_file_chooser_get_uri<%s>\n", uri)))
  return uri;
}

gboolean gtk_file_chooser_set_uri (GtkFileChooser *chooser, const char *uri)
{
  return gtk_file_chooser_set_filename (chooser, from_uri (uri));
}

GSList *gtk_file_chooser_get_uris (GtkFileChooser *chooser)
{
  DEBUG ((printf ("gtk_file_chooser_get_uris\n")))
  return to_uris (gtk_file_chooser_get_filenames (chooser));
}

gboolean gtk_file_chooser_set_current_folder_uri(GtkFileChooser *chooser, const gchar *uri)
{
  return gtk_file_chooser_set_current_folder (chooser, from_uri (uri));
}

gchar *gtk_file_chooser_get_current_folder_uri (GtkFileChooser *chooser)
{
  DEBUG ((printf ("gtk_file_chooser_get_current_folder\n")))
  return to_uri (gtk_file_chooser_get_current_folder (chooser));
}

/////////////////////////////////////////////////////////////////////////////

static gboolean disable_cb (gpointer data)
{
  ((void (*)(void *, const char *)) dlsym (RTLD_NEXT, "gtk_file_chooser_set_filename"))(GTK_FILE_CHOOSER (data), "/tmp/gtkbfc/empty");

  GtkWidget *toplevel = gtk_widget_get_toplevel (data);
  if (toplevel) gtk_widget_set_sensitive (toplevel, 0);

  gtk_widget_unref (data);
  return FALSE;
}

static gboolean read_result (GIOChannel *source, GIOCondition condition, gpointer data)
{
  g_free (last_path); last_path = 0;

  g_io_channel_read_to_end (source, &last_path, 0, 0);
  g_io_channel_unref (source);

  GtkWidget *toplevel = gtk_widget_get_toplevel (data);
  if (toplevel) gtk_widget_set_sensitive (toplevel, 1);

  if (!last_path || *last_path != '/')
    {
      DEBUG ((printf ("read_result cancel\n")))
      g_free (last_path);
      last_path = 0;

      if (toplevel)
        {
          GdkEvent ev = { GDK_DELETE };
          g_signal_emit_by_name (toplevel, "delete-event", &ev);
        }
    }
  else
    {
      DEBUG ((printf ("read_result<%s>\n", last_path)));
      gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (data), FALSE);
      g_signal_emit_by_name (GTK_FILE_CHOOSER (data), "file-activated");
    }

  gtk_widget_unref (data);
  return FALSE;
}

static gboolean map_cb (GtkWidget *widget, gpointer user_data) 
{
  gchar *argv[10];
  gboolean save = 0;
  char xid [40];

  argv [0] = helper;

  switch (gtk_file_chooser_get_action (GTK_FILE_CHOOSER (widget)))
    {
      case GTK_FILE_CHOOSER_ACTION_SAVE:          argv [1] = "save "; save = 1;break;
      case GTK_FILE_CHOOSER_ACTION_OPEN:          argv [1] = "load "; break;
      case GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER: argv [1] = "mkdir "; save = 1; break;
      case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER: argv [1] = "cd "; break;
      default: argv [1] = "<unknown action> "; break;
    }

  GdkWindow *w = gtk_widget_get_parent_window (GTK_WIDGET (widget));

  snprintf (xid, 40, "%lu", (unsigned long)(w ? gdk_x11_drawable_get_xid (w) : 0));
  argv [2] = xid;
  argv [3] = last_filename; // may be 0
  argv [4] = 0;

  mkdir ("/tmp/gtkbfc", 0777);
  close (open ("/tmp/gtkbfc/empty", O_WRONLY|O_CREAT|O_TRUNC, 0666));

  ((void (*)(void *, const char *)) dlsym (RTLD_NEXT, "gtk_file_chooser_set_filename"))(GTK_FILE_CHOOSER (widget), "/tmp/gtkbfc/empty");

  gint stdout_fd = -1;

  gtk_widget_ref (widget); disable_cb (widget);

  if (g_spawn_async_with_pipes (0, argv, 0, 0, 0, 0, 0, 0, &stdout_fd, 0, 0))
    {
      GIOChannel *channel = g_io_channel_unix_new (stdout_fd);

      gtk_widget_ref (widget); g_io_add_watch (channel, G_IO_IN  | G_IO_ERR | G_IO_HUP, read_result, widget);

      // disable again in an idle handler, to get around race conditions
      gtk_widget_ref (widget); g_idle_add (disable_cb, widget);
    }
}

static GObject *new_constructor (GType type, guint n_construct_properties, GObjectConstructParam *construct_properties)
{
  GObject *ob = old_constructor (type, n_construct_properties, construct_properties);
  g_signal_connect_after (ob, "map", G_CALLBACK (map_cb), 0);
  return ob;
}

__attribute__ ((constructor))
static void init ()
{
  g_type_init ();

  GType type = GTK_TYPE_FILE_CHOOSER_WIDGET;
  GObjectClass *klass = g_type_class_ref (type);

  old_constructor = klass->constructor;
  klass->constructor = new_constructor;
}

